==> coventry_0.5.3.orig.tar.gz: coventry-0.5.3/server/network.cpp <==
/*
 * Copyright (C) 2020-2021 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "main.hpp"
#include "network.hpp"

#include <strings.hpp>

#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <ifaddrs.h>
#include <netinet/in.h>
#include <net/if.h>
#include <arpa/inet.h>

using namespace server;

auto net::hostname() -> std::string
{
    char buf[256];
    auto result = gethostname(buf, sizeof(buf));
    if(result != 0)
        throw std::invalid_argument("cannot get hostname");
    buf[sizeof(buf) - 1] = 0;
    return buf;
}

auto net::to_string(const std::string& address, int port) -> std::string
{
    auto host = address;

    if(address.empty() && ipv6)
        host = "[::]";
    else if(address.empty())
        host = "0.0.0.0";
    else if(address.find_first_of(':') != std::string::npos && address[0] != '[')
        host = "[" + address + "]";

    if(!port)
        return host;

    return host + ":" + std::to_string(port);
}

void net::fqdn(std::string& host)
{
    struct addrinfo hints, *info; // NOLINT
    auto hostname = host.c_str();
    if(!hostname || !*hostname) {
        host = net::hostname();
        hostname = host.c_str();
    }

    memset(&hints, 0, sizeof hints);
    hints.ai_family = AF_UNSPEC;    // either IPV4 or IPV6
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_CANONNAME;

    if (getaddrinfo(hostname, nullptr, &hints, &info) != 0)
        return;

    if(info) {
        host = info->ai_canonname;
        freeaddrinfo(info);
    }
}

auto net::size(const struct sockaddr *addr) -> socklen_t
{
    switch(addr->sa_family) {
    case AF_INET:
        return sizeof(struct sockaddr_in);
    case AF_INET6:
        return sizeof(struct sockaddr_in6);
    default:
        return socklen_t(0);
    }
}

auto net::to_address(const std::string& name) -> std::string
{
    using namespace std::literals::string_literals;
    struct addrinfo hints{}, *info{nullptr};

    memset(&hints, 0, sizeof(hints));
    hints.ai_family = ipv6 ? AF_INET6 : AF_INET;
    hints.ai_flags |= AI_ADDRCONFIG;
    if(getaddrinfo(name.c_str(), nullptr, &hints, &info) != 0)
        return "";

    if(info && info->ai_addr) {
        char buf[INET6_ADDRSTRLEN + 1];
        auto hp = info->ai_addr;
        auto sin = reinterpret_cast<struct sockaddr_in *>(hp);
        auto sin6 = reinterpret_cast<struct sockaddr_in6 *>(hp);
        auto cp = "";
        auto family = hp->sa_family;

        switch(family) {
        case AF_INET:
            cp = inet_ntop(AF_INET, &sin->sin_addr, buf, sizeof(buf));
            break;
        case AF_INET6:
            cp = inet_ntop(AF_INET6, &sin6->sin6_addr, buf, sizeof(buf));
            break;
        default:
            break;
        }
        freeaddrinfo(info);
        if(!cp)
            return "";

        if(family == AF_INET6)
            return "["s + cp + "]"s;
        return cp;
    }
    return "";
}

auto net::if_address(const std::string& iface) -> std::string
{
    if((iface.empty() || iface == "*") && ipv6)
        return "::";

    if(iface.empty() || iface == "*")
        return "*";

    if((iface == "localhost" || iface == "loopback") && ipv6)
        return "::1";

    if((iface == "localhost" || iface == "loopback"))
        return "127.0.0.1";

    if(!ipv6 && iface.find_first_of(':') != std::string::npos)
        throw std::invalid_argument("bad interface address");

    if(iface.find_first_of(".:") != std::string::npos)
        return iface;

    int family = AF_INET;
    if(ipv6)
        family = AF_INET6;

    void *sin = nullptr;
    struct ifaddrs *ifa = nullptr, *entry = nullptr;
    getifaddrs(&ifa);
    for(entry = ifa; entry != nullptr; entry = entry->ifa_next) {
        if(entry->ifa_addr->sa_family == family && entry->ifa_name && !strcmp(iface.c_str(), entry->ifa_name)) {
            if(ipv6)
                sin = &reinterpret_cast<struct sockaddr_in6 *>(entry->ifa_addr)->sin6_addr;
            else
                sin = &reinterpret_cast<struct sockaddr_in *>(entry->ifa_addr)->sin_addr;
            break;
        }
    }

    if(!sin) {
        freeifaddrs(ifa);
        throw std::invalid_argument("bad interface name");
    }

    char buf[INET6_ADDRSTRLEN + 1];
    auto result = inet_ntop(family, sin, buf, sizeof(buf));
    freeifaddrs(ifa);
    return result;
}

auto net::uri_schema(const std::string& id) -> std::string
{
    if(begins_with(id, "sip:"))
        return "sip:";
    return {};
}

auto net::uri_userid(const std::string& id) -> std::string
{
    auto user = id;
    if(begins_with(id, "sip:"))
        user = id.substr(4);

    auto pos = user.find_last_of('@');
    if(pos != std::string::npos)
        return user.substr(0, pos);
    return user;
}

auto net::uri_server(const std::string& id) -> std::string
{
    auto server = id;
    if(begins_with(id, "sip:"))
        server = id.substr(4);

    auto pos = server.find_last_of('@');
    if(pos != std::string::npos)
        return server.substr(pos + 1);
    return server;
}

auto net::uri_host(const std::string& id) -> std::string
{
    auto server = id;

    auto pos = id.find_last_of('@');
    if(pos != std::string::npos)
        server = id.substr(pos + 1);
    else {
        if(begins_with(id, "sip:"))
            server = id.substr(4);
    }

    auto ipv6 = server.find_last_of(']');
    pos = server.find_last_of(':');

    if((pos != std::string::npos) && (ipv6 == std::string::npos || ipv6 < pos))
        return server.substr(0, pos);

    return server;
}

auto net::uri_address(const std::string& id) -> std::string
{
    auto host = uri_host(id);
    auto addr = to_address(host);

    if(host.empty() || addr.empty())
        return host;

    return addr;
}

auto net::uri_port(const std::string& id) -> uint16_t
{
    auto server = id;
    auto port = uint16_t(5060);

    if(begins_with(id, "sip:"))
        server = id.substr(4);

    auto ipv6 = server.find_last_of(']');
    auto pos = server.find_last_of(':');

    if(pos == std::string::npos)
        return port;

    if(ipv6 != std::string::npos && ipv6 > pos)
        return port;

    return std::stoi(server.substr(pos + 1));
}

auto net::set_user(const std::string& uri, const std::string& user) -> std::string
{
    auto schema = uri_schema(uri);
    auto server = uri_server(uri);

    auto ipv6 = server.find_last_of(']');
    auto pos = server.find_last_of(':');

    if((pos != std::string::npos) && (ipv6 == std::string::npos || ipv6 < pos)) {
        auto port = std::stoi(server.substr(pos + 1));
        return make_uri(schema, user, server.substr(0, pos), port);               }
    return make_uri(schema, user, server);
}

auto net::make_uri(const std::string& schema, const std::string& host, uint16_t port) -> std::string
{
    if(!port)
        return schema + host;

    return schema + host + ":" + std::to_string(port);
}

auto net::make_uri(const std::string& schema, const std::string& user, const std::string& host, uint16_t port) -> std::string
{
    if(!port)
        return schema + user + "@" + host;

    return schema + user + "@" + host + ":" + std::to_string(port);
}

auto net::store(struct sockaddr_storage& addr, const std::string& host, uint16_t port) -> bool
{
    auto service = std::to_string(port);
    struct addrinfo hints{}, *list = nullptr;
    memset(&hints, 0, sizeof(hints));
    memset(&addr, 0, sizeof(addr));
    hints.ai_family = ipv6 ? AF_INET6 : AF_INET;
    if(getaddrinfo(host.c_str(), service.c_str(), &hints, &list))
        list = nullptr;

    if(list) {
        size_t size = ipv6 ? sizeof(struct sockaddr_in6) : sizeof(struct sockaddr_in);
        memcpy(&addr, list->ai_addr, size); // FlawFinder: size valid
        freeaddrinfo(list);
        return true;
    }
    return false;
}

auto net::is_ipv4(const char *cp) -> bool
{
    while(cp && *cp) {
        if(*cp != '.' && !isdigit(*cp))
            return false;
        ++cp;
    }
    return true;
}

auto net::is_ipv6(const char *cp) -> bool
{
    return strchr(cp, ':') != nullptr;
}

==> coventry_0.7.6.orig.tar.gz: coventry-0.7.6/server/network.cpp <==
/*
 * Copyright (C) 2020-2023 David Sugar <tychosoft@gmail.com>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "main.hpp"
#include "network.hpp"

#include <strings.hpp>

#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <ifaddrs.h>
#include <netinet/in.h>
#include <net/if.h>
#include <arpa/inet.h>

using namespace server;

auto net::hostname() -> std::string {
    char buf[256];
    auto result = gethostname(buf, sizeof(buf));
    if(result != 0)
        throw std::invalid_argument("cannot get hostname");
    buf[sizeof(buf) - 1] = 0;
    return buf;
}

auto net::interface() -> std::string {
    const auto addr = "4.2.2.1";
    int sock{-1};
    socklen_t len{0};
#ifdef  AF_NET6
    if(ipv6)
        sock = socket(AF_INET6, SOCK_DGRAM, 0);
    else
#endif
    sock = socket(AF_INET, SOCK_DGRAM, 0);
    if(sock < 0)
        throw std::runtime_error("cannot create socket");

#ifdef  AF_INET6
    if(ipv6) {
        char buf[INET6_ADDRSTRLEN];
        struct sockaddr_in6 in6{};
        struct in_addr in4{};
        len = sizeof(in6);
        std::memset(&in6, 0, sizeof(in6));
        inet_pton(AF_INET, addr, &(in4.s_addr));
        memcpy(reinterpret_cast<char *>(&in6.sin6_addr) + 12, &in4.s_addr, 4); // FlawFinder: ignore
        //in6.sin6_addr.s6_addr32[3] = in4.s_addr;
        in6.sin6_family = AF_INET6;
        in6.sin6_port = htons(80);  // Use any available port
        if(connect(sock, reinterpret_cast<struct sockaddr *>(&in6), len) < -1)
            throw std::runtime_error("cannot connect socket");

        getsockname(sock, reinterpret_cast<struct sockaddr *>(&in6), &len);
        ::close(sock);
        if(nullptr == inet_ntop(AF_INET6, &in6, buf, sizeof(buf)))
            throw std::runtime_error("cannot connect socket");

        return buf;
    }
#endif
    struct sockaddr_in in{};
    len = sizeof(in);
    std::memset(&in, 0, sizeof(in));
    in.sin_family = AF_INET;
    inet_pton(AF_INET, addr, &(in.sin_addr.s_addr));
    in.sin_port = htons(80);  // Use any available port
    if(connect(sock, reinterpret_cast<struct sockaddr *>(&in), len) < -1)
        throw std::runtime_error("cannot connect socket");

    getsockname(sock, reinterpret_cast<struct sockaddr *>(&in), &len);
    ::close(sock);
    return inet_ntoa(in.sin_addr);
}

auto net::to_string(const std::string& address, int port) -> std::string {
    auto host = address;

    if(address.empty() && ipv6)
        host = "[::]";
    else if(address.empty())
        host = "0.0.0.0";
    else if(address.find_first_of(':') != std::string::npos && address[0] != '[')
        host = "[" + address + "]";

    if(!port)
        return host;

    return host + ":" + std::to_string(port);
}

void net::fqdn(std::string& host) {
    struct addrinfo hints, *info; // NOLINT
    auto hostid = host.c_str();
    if(!*hostid) {
        host = net::hostname();
        hostid = host.c_str();
    }

    memset(&hints, 0, sizeof hints);
    hints.ai_family = AF_UNSPEC;    // either IPV4 or IPV6
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_CANONNAME;

    if(getaddrinfo(hostid, nullptr, &hints, &info) != 0)
        return;

    if(info) {
        host = info->ai_canonname;
        freeaddrinfo(info);
    }
}

auto net::size(const struct sockaddr *addr) -> socklen_t {
    switch(addr->sa_family) {
    case AF_INET:
        return sizeof(struct sockaddr_in);
    case AF_INET6:
        return sizeof(struct sockaddr_in6);
    default:
        return socklen_t(0);
    }
}

auto net::to_address(const std::string& name) -> std::string {
    using namespace std::literals::string_literals;
    struct addrinfo hints{}, *info{nullptr};

    memset(&hints, 0, sizeof(hints));
    hints.ai_family = ipv6 ? AF_INET6 : AF_INET;
    hints.ai_flags |= AI_ADDRCONFIG;
    if(getaddrinfo(name.c_str(), nullptr, &hints, &info) != 0)
        return "";

    if(info && info->ai_addr) {
        char buf[INET6_ADDRSTRLEN + 1];
        auto hp = info->ai_addr;
        auto sin = reinterpret_cast<struct sockaddr_in *>(hp);
        auto sin6 = reinterpret_cast<struct sockaddr_in6 *>(hp);
        auto cp = "";
        auto family = hp->sa_family;

        switch(family) {
        case AF_INET:
            cp = inet_ntop(AF_INET, &sin->sin_addr, buf, sizeof(buf));
            break;
        case AF_INET6:
            cp = inet_ntop(AF_INET6, &sin6->sin6_addr, buf, sizeof(buf));
            break;
        default:
            break;
        }
        freeaddrinfo(info);
        if(!cp)
            return "";

        if(family == AF_INET6)
            return "["s + cp + "]"s;
        return cp;
    }
    return "";
}

auto net::if_address(const std::string& iface) -> std::string {
    if((iface.empty() || iface == "*") && ipv6)
        return "::";

    if(iface.empty() || iface == "*")
        return "*";

    if((iface == "localhost" || iface == "loopback") && ipv6)
        return "::1";

    if((iface == "localhost" || iface == "loopback"))
        return "127.0.0.1";

    if(!ipv6 && iface.find_first_of(':') != std::string::npos)
        throw std::invalid_argument("bad interface address");

    if(iface.find_first_of(".:") != std::string::npos)
        return iface;

    int family = AF_INET;
    if(ipv6)
        family = AF_INET6;

    void *sin = nullptr;
    struct ifaddrs *ifa = nullptr, *entry = nullptr;
    getifaddrs(&ifa);
    for(entry = ifa; entry != nullptr; entry = entry->ifa_next) {
        if(entry->ifa_addr->sa_family == family && entry->ifa_name && !strcmp(iface.c_str(), entry->ifa_name)) {
            if(ipv6)
                sin = &reinterpret_cast<struct sockaddr_in6 *>(entry->ifa_addr)->sin6_addr;
            else
                sin = &reinterpret_cast<struct sockaddr_in *>(entry->ifa_addr)->sin_addr;
            break;
        }
    }

    if(!sin) {
        freeifaddrs(ifa);
        throw std::invalid_argument("bad interface name");
    }

    char buf[INET6_ADDRSTRLEN + 1];
    auto result = inet_ntop(family, sin, buf, sizeof(buf));
    freeifaddrs(ifa);
    return result;
}

auto net::uri_schema(const std::string& id) -> std::string {
    if(begins_with(id, "sip:"))
        return "sip:";
    return {};
}

auto net::uri_userid(const std::string& id) -> std::string {
    auto user = id;
    if(begins_with(id, "sip:"))
        user = id.substr(4);

    auto pos = user.find_last_of('@');
    if(pos != std::string::npos)
        return user.substr(0, pos);
    return user;
}

auto net::uri_server(const std::string& id) -> std::string {
    auto server = id;
    if(begins_with(id, "sip:"))
        server = id.substr(4);

    auto pos = server.find_last_of('@');
    if(pos != std::string::npos)
        return server.substr(pos + 1);
    return server;
}

auto net::uri_host(const std::string& id) -> std::string {
    auto server = id;

    auto pos = id.find_last_of('@');
    if(pos != std::string::npos)
        server = id.substr(pos + 1);
    else {
        if(begins_with(id, "sip:"))
            server = id.substr(4);
    }

    auto inet6 = server.find_last_of(']');
    pos = server.find_last_of(':');

    if((pos != std::string::npos) && (inet6 == std::string::npos || inet6 < pos))
        return server.substr(0, pos);

    return server;
}

auto net::uri_address(const std::string& id) -> std::string {
    auto host = uri_host(id);
    auto addr = to_address(host);

    if(host.empty() || addr.empty())
        return host;

    return addr;
}

auto net::uri_port(const std::string& id) -> uint16_t {
    auto server = id;
    auto port = uint16_t(5060);

    if(begins_with(id, "sip:"))
        server = id.substr(4);

    auto inet6 = server.find_last_of(']');
    auto pos = server.find_last_of(':');

    if(pos == std::string::npos)
        return port;

    if(inet6 != std::string::npos && inet6 > pos)
        return port;

    return std::stoi(server.substr(pos + 1));
}

auto net::set_user(const std::string& uri, const std::string& user) -> std::string {
    auto schema = uri_schema(uri);
    auto server = uri_server(uri);

    auto inet6 = server.find_last_of(']');
    auto pos = server.find_last_of(':');

    if((pos != std::string::npos) && (inet6 == std::string::npos || inet6 < pos)) {
        auto port = std::stoi(server.substr(pos + 1));
        return make_uri(schema, user, server.substr(0, pos), port);               }
    return make_uri(schema, user, server);
}

auto net::make_uri(const std::string& schema, const std::string& host, uint16_t port) -> std::string {
    if(!port)
        return schema + host;

    return schema + host + ":" + std::to_string(port);
}

auto net::make_uri(const std::string& schema, const std::string& user, const std::string& host, uint16_t port) -> std::string {
    if(!port)
        return schema + user + "@" + host;

    return schema + user + "@" + host + ":" + std::to_string(port);
}

auto net::store(struct sockaddr_storage& addr, const std::string& host, uint16_t port) -> bool {
    auto service = std::to_string(port);
    struct addrinfo hints{}, *list = nullptr;
    memset(&hints, 0, sizeof(hints));
    memset(&addr, 0, sizeof(addr));
    hints.ai_family = ipv6 ? AF_INET6 : AF_INET;
    if(getaddrinfo(host.c_str(), service.c_str(), &hints, &list))
        list = nullptr;

    if(list) {
        auto asize = ipv6 ? sizeof(struct sockaddr_in6) : sizeof(struct sockaddr_in);
        memcpy(&addr, list->ai_addr, asize); // FlawFinder: size valid
        freeaddrinfo(list);
        return true;
    }
    return false;
}

auto net::is_ipv4(const char *cp) -> bool {
    while(cp && *cp) {
        if(*cp != '.' && !isdigit(*cp))
            return false;
        ++cp;
    }
    return true;
}

auto net::is_ipv6(const char *cp) -> bool {
    return strchr(cp, ':') != nullptr;
}
